from typing import AsyncIterable

from aiomysql import connect
from aett.eventstore import IManagePersistenceAsync, TopicMap, COMMITS, SNAPSHOTS, Commit
from aett.storage.synchronous.postgresql import _item_to_commit


class AsyncPersistenceManagement(IManagePersistenceAsync):
    def __init__(
            self,
            host: str,
            user: str,
            password: str,
            database: str,
            topic_map: TopicMap,
            port: int = 3306,
            commits_table_name: str = COMMITS,
            snapshots_table_name: str = SNAPSHOTS,
    ):
        self.host: str = host
        self._user: str = user
        self._password: str = password
        self._database: str = database
        self._port: int = port
        self._topic_map = topic_map
        self._commits_table_name = commits_table_name
        self._snapshots_table_name = snapshots_table_name

    async def initialize(self):
        try:
            stmts = [stmt.strip() for stmt in f"""CREATE TABLE IF NOT EXISTS {self._commits_table_name}
                     (
                         TenantId varchar(36) charset utf8 NOT NULL,
                         StreamId varchar(36) charset utf8 NOT NULL,
                         StreamIdOriginal varchar(1000) charset utf8 NOT NULL,
                         StreamRevision int NOT NULL CHECK (StreamRevision > 0),
                         Items tinyint NOT NULL CHECK (Items > 0),
                         CommitId binary(16) NOT NULL,
                         CommitSequence int NOT NULL CHECK (CommitSequence > 0),
                         CommitStamp TIMESTAMP NOT NULL,
                         CheckpointNumber bigint AUTO_INCREMENT,
                         Headers blob NULL,
                         Payload mediumblob NOT NULL,
                         PRIMARY KEY (CheckpointNumber)
                     );
                     CREATE UNIQUE INDEX IX_Commits ON {self._commits_table_name} (TenantId, StreamId, CommitSequence);
                     CREATE UNIQUE INDEX IX_Commits_CommitId ON {self._commits_table_name} (TenantId, StreamId, CommitId);
                     CREATE UNIQUE INDEX IX_Commits_Revisions ON {self._commits_table_name} (TenantId, StreamId, StreamRevision, Items);
                     CREATE INDEX IX_Commits_Stamp ON {self._commits_table_name} (CommitStamp);

                     CREATE TABLE IF NOT EXISTS {self._snapshots_table_name}
                     (
                         TenantId varchar(64) charset utf8 NOT NULL,
                         StreamId varchar(64) charset utf8 NOT NULL,
                         StreamRevision int NOT NULL CHECK (StreamRevision > 0),
                         CommitSequence int NOT NULL CHECK (CommitSequence > 0),
                         Payload blob NOT NULL,
                         Headers blob NOT NULL,
                         CONSTRAINT PK_Snapshots PRIMARY KEY (TenantId, StreamId, StreamRevision)
                     );""".split(';') if stmt.strip()]
            connection = await connect(
                host=self.host,
                user=self._user,
                password=self._password,
                db=self._database,
                port=self._port,
                autocommit=True)
            cur = await  connection.cursor()
            for stmt in stmts:
                await cur.execute(query=stmt)
            await cur.close()
            await connection.commit()
            await connection.close()
        except Exception as e:
            print(e)
            pass

    async def drop(self):
        connection = await connect(host=self.host,
                                   user=self._user,
                                   password=self._password,
                                   db=self._database,
                                   port=self._port,
                                   autocommit=True)
        c = await connection.cursor()
        await c.execute(
            f"""DROP TABLE {self._snapshots_table_name};"""
        )
        await c.execute(
            f"""DROP TABLE {self._commits_table_name};"""
        )
        await c.close()
        await connection.commit()
        await connection.close()

    async def purge(self, tenant_id: str):
        async with connect(host=self.host,
                           user=self._user,
                           password=self._password,
                           db=self._database,
                           port=self._port,
                           autocommit=True) as connection:
            async with connection.cursor() as c:
                c.execute(
                    f"""DELETE FROM {self._commits_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )
                c.execute(
                    f"""DELETE FROM {self._snapshots_table_name} WHERE TenantId = %s;""",
                    tenant_id,
                )
            connection.commit()

    async def get_from(self, checkpoint: int) -> AsyncIterable[Commit]:
        async with connect(host=self.host,
                           user=self._user,
                           password=self._password,
                           db=self._database,
                           port=self._port,
                           autocommit=True) as connection:
            async with connection.cursor() as cur:
                await cur.execute(
                    f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                                  FROM {self._commits_table_name}
                                 WHERE CommitStamp >= %s
                                 ORDER BY CheckpointNumber;""",
                    (checkpoint,),
                )
                fetchall = await cur.fetchall()
                for doc in fetchall:
                    yield _item_to_commit(doc, self._topic_map)



from typing import TypeVar

from aett.eventstore import BaseEvent, Memento
from aett.eventstore.base_command import BaseCommand

TMemento = TypeVar("TMemento", bound=Memento)
TCommand = TypeVar("TCommand", bound=BaseCommand)

TUncommitted = TypeVar("TUncommitted", bound=BaseEvent)
TCommitted = TypeVar("TCommitted", bound=BaseEvent)

UNDISPATCHEDMESSAGES = "UndispatchedMessage"


import datetime
import typing
from uuid import UUID

from aiomysql import connect, IntegrityError
from pydantic_core import to_json, from_json

from aett.domain import (
    ConflictDetector,
    DuplicateCommitException,
    ConflictingCommitException,
    NonConflictingCommitException,
)
from aett.eventstore import (
    ICommitEventsAsync,
    TopicMap,
    COMMITS,
    MAX_INT,
    Commit,
    EventMessage, BaseEvent,
)
from aett.storage.asynchronous.mysql import _item_to_commit


class AsyncCommitStore(ICommitEventsAsync):
    def __init__(
            self,
            host: str,
            user: str,
            password: str,
            database: str,
            topic_map: TopicMap,
            port: int = 3306,
            conflict_detector: ConflictDetector | None = None,
            table_name=COMMITS,
    ):
        self._topic_map = topic_map if topic_map else TopicMap()
        self._host: str = host
        self._user: str = user
        self._password: str = password
        self._database: str = database
        self._port: int = port
        self._conflict_detector = (
            conflict_detector if conflict_detector is not None else ConflictDetector()
        )
        self._table_name = table_name

    async def get(
            self,
            tenant_id: str,
            stream_id: str,
            min_revision: int = 0,
            max_revision: int = MAX_INT,
    ) -> typing.AsyncIterable[Commit]:
        max_revision = MAX_INT if max_revision >= MAX_INT else max_revision + 1
        min_revision = 0 if min_revision < 0 else min_revision
        connection = await connect(host=self._host,
                                   user=self._user,
                                   password=self._password,
                                   db=self._database,
                                   port=self._port,
                                   autocommit=True)
        cur = await connection.cursor()
        await cur.execute(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
          FROM {self._table_name}
         WHERE TenantId = %s
           AND StreamId = %s
           AND StreamRevision >= %s
           AND (StreamRevision - Items) < %s
           AND CommitSequence > %s
         ORDER BY CommitSequence;""",
            (tenant_id, stream_id, min_revision, max_revision, 0),
        )
        fetchall = await cur.fetchall()
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)
        await cur.close()
        connection.close()

    async def get_to(
            self,
            tenant_id: str,
            stream_id: str,
            max_time: datetime.datetime = datetime.datetime.max,
    ) -> typing.AsyncIterable[Commit]:
        connection = await connect(host=self._host,
                                   user=self._user,
                                   password=self._password,
                                   db=self._database,
                                   port=self._port,
                                   autocommit=True)
        cur = await connection.cursor()
        await cur.execute(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                  FROM {self._table_name}
                 WHERE TenantId = %s
                   AND StreamId = %s
                   AND CommitStamp <= %s
                 ORDER BY CommitSequence;""",
            (tenant_id, stream_id, max_time),
        )
        fetchall = await cur.fetchall()
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)
        await cur.close()
        connection.close()

    async def get_all_to(
            self, tenant_id: str, max_time: datetime.datetime = datetime.datetime.max
    ) -> typing.AsyncIterable[Commit]:
        connection = await connect(host=self._host,
                                   user=self._user,
                                   password=self._password,
                                   db=self._database,
                                   port=self._port,
                                   autocommit=True)
        cur = await connection.cursor()
        await cur.execute(
            f"""SELECT TenantId, StreamId, StreamIdOriginal, StreamRevision, CommitId, CommitSequence, CommitStamp,  CheckpointNumber, Headers, Payload
                              FROM {self._table_name}
                             WHERE TenantId = %s
                               AND CommitStamp <= %s
                             ORDER BY CheckpointNumber;""",
            (tenant_id, max_time),
        )
        fetchall = cur.fetchall()
        for doc in fetchall:
            yield _item_to_commit(doc, self._topic_map)
        await cur.close()
        connection.close()

    async def commit(self, commit: Commit) -> Commit:
        try:
            async with connect(
                    host=self._host,
                    user=self._user,
                    password=self._password,
                    db=self._database,
                    port=self._port,
                    autocommit=True) as connection:
                async with connection.cursor() as cur:
                    json = to_json([e.to_json() for e in commit.events])
                    await cur.execute(
                        f"""INSERT
                      INTO {self._table_name}
                         ( TenantId, StreamId, StreamIdOriginal, CommitId, CommitSequence, StreamRevision, Items, CommitStamp, Headers, Payload )
                    VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s);""",
                        (
                            commit.tenant_id,
                            commit.stream_id,
                            commit.stream_id,
                            commit.commit_id.bytes,
                            commit.commit_sequence,
                            commit.stream_revision,
                            len(commit.events),
                            commit.commit_stamp,
                            to_json(commit.headers),
                            json,
                        ),
                    )
                    checkpoint_number = cur.lastrowid
                    return Commit(
                        tenant_id=commit.tenant_id,
                        stream_id=commit.stream_id,
                        stream_revision=commit.stream_revision,
                        commit_id=commit.commit_id,
                        commit_sequence=commit.commit_sequence,
                        commit_stamp=commit.commit_stamp,
                        headers=commit.headers,
                        events=commit.events,
                        checkpoint_token=checkpoint_number,
                    )
        except IntegrityError:
            if await self._detect_duplicate(
                    commit.commit_id, commit.tenant_id, commit.stream_id
            ):
                raise DuplicateCommitException(
                    f"Commit {commit.commit_id} already exists in stream {commit.stream_id}"
                )
            else:
                conflicts, revision = await self._detect_conflicts(commit=commit)
                if conflicts:
                    raise ConflictingCommitException(
                        f"Conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )
                else:
                    raise NonConflictingCommitException(
                        f"Non-conflicting version conflict detected in stream {commit.stream_id} with revision {commit.stream_revision}"
                    )
        except Exception as e:
            raise Exception(f"Failed to commit {commit.commit_id} with error {e}")

    async def _detect_duplicate(
            self, commit_id: UUID, tenant_id: str, stream_id: str
    ) -> bool:
        try:
            connection = await connect(host=self._host,
                                       user=self._user,
                                       password=self._password,
                                       db=self._database,
                                       port=self._port,
                                       autocommit=True)
            cur = await connection.cursor()
            await cur.execute(
                f"""SELECT COUNT(*)
              FROM {self._table_name}
             WHERE TenantId = %s
               AND StreamId = %s
               AND CommitId = %s;""",
                (tenant_id, stream_id, str(commit_id)),
            )
            result = await  cur.fetchone()
            await cur.close()
            connection.close()
            return result[0] > 0
        except Exception as e:
            raise Exception(
                f"Failed to detect duplicate commit {commit_id} with error {e}"
            )

    async def _detect_conflicts(self, commit: Commit) -> typing.Tuple[bool, int]:
        connection = await connect(host=self._host,
                                   user=self._user,
                                   password=self._password,
                                   db=self._database,
                                   port=self._port,
                                   autocommit=True)
        cur = await connection.cursor()
        await cur.execute(
            f"""SELECT StreamRevision, Payload
                              FROM {self._table_name}
                             WHERE TenantId = %s
                               AND StreamId = %s
                               AND StreamRevision <= %s
                             ORDER BY CommitSequence;""",
            (commit.tenant_id, commit.stream_id, commit.stream_revision),
        )
        fetchall = await cur.fetchall()
        latest_revision = 0
        for doc in fetchall:
            events = [
                EventMessage.from_dict(e, self._topic_map)
                for e in from_json(doc[1])
            ]
            uncommitted_events = list(map(self._get_body, commit.events))
            committed_events = list(map(self._get_body, events))
            if self._conflict_detector.conflicts_with(
                    uncommitted_events, committed_events
            ):
                return True, -1
            if doc[0] > latest_revision:
                latest_revision = int(doc[0])
        return False, latest_revision

    @staticmethod
    def _get_body(e: EventMessage) -> BaseEvent:
        return e.body
